/* OpenCP Module Player
 * copyright (c) '94-'05 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * pak binfile (handles files in opencp's CP.PAK file)
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -fd981207   Felix Domke <tmbinc@gmx.net>
 *    -edited for new binfile
 *  -ss040601   Stian Skjelstad <stian@nixia.no>
 *    -rewritten to use FILE * instead of binfile
 */

#include "config.h"
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "types.h"
#include "binfpak.h"
#include "stuff/err.h"
#include "boot/psetting.h"

struct __attribute__((packed)) packdirentry
{
	char name[0x38];
	uint32_t off;
	uint32_t len;
};

static int nfiles;
static struct packdirentry *dir;
static FILE *packfile=NULL;

FILE *pakfOpen(char *name)
{
	int i;
	FILE *retval;
	char *tmp;
	if (!packfile)
		return NULL;
	for (i=0;i<nfiles;i++)
		if (!strcasecmp(name, dir[i].name))
			break;
	if (i==nfiles)
		return NULL;
	fseek(packfile, dir[i].off, SEEK_SET);
	if  ( ! ( retval = tmpfile () ) )
	{
		perror ("tempfile()");
		return NULL;
	}
	if (fcntl(fileno(retval), F_SETFD, FD_CLOEXEC))
		perror("fcntl(fileno, F_SETFD, FD_CLOEXEC)");
	tmp=malloc(dir[i].len);
	if (fread(tmp, dir[i].len, 1, packfile)!=1)
	{
		perror ( __FILE__ " pakfOpen(): fread() failed\n");
		free(tmp);
		fclose(retval);
		return NULL;
	}
	if (fwrite(tmp, dir[i].len, 1, retval)!=1)
	{
		perror ( __FILE__ " pakfOpen(): fwrite() failed\n");
		free(tmp);
		fclose(retval);
		return NULL;
	}
	free(tmp);
	fseek(retval, 0, SEEK_SET);
	return retval;
}

int pakfInit(void)
{
	char path[PATH_MAX+1];
	uint32_t tmp;
	int o;
	int i,j;

	snprintf(path, sizeof(path), "%s%s", cfDataDir, "ocp.pak");
	nfiles=0;
	dir=0;
	if (!(packfile=fopen(path, "r")))
	{
		perror(__FILE__ " fopen()");
		goto error_out;
	}
	if (!fread(&tmp, sizeof(uint32_t), 1, packfile))
	{
		perror (__FILE__ " fread()");
		goto error_out;
	}
	if ( tmp != uint32_little (0x4B434150) )
		goto error_out;
	if (!fread(&o, sizeof(o), 1, packfile))
		goto error_out;
	o = uint32_little (o);
	if (!fread(&nfiles, sizeof(nfiles), 1, packfile))
	{
		perror (__FILE__ " fread()");
		goto error_out;
	}
	nfiles = uint32_little (nfiles);
	nfiles/=0x40;
	if (!nfiles)
		goto error_out;
	dir=calloc(sizeof(dir[0]), nfiles);
	fseek(packfile, o, SEEK_SET);
	if (!fread(dir, sizeof(dir[0]), nfiles, packfile))
	{
		perror (__FILE__ " fread()");
		goto error_out;
	}
	for (i=0; i<nfiles; i++)
	{
		dir[i].off = uint32_little (dir[i].off);
		dir[i].len = uint32_little (dir[i].len);
		for (j=0; j<0x38; j++)
			if (dir[i].name[j]=='\\')   /* we are hardcore unix folks */
				dir[i].name[j]='/';

	}
	if (fcntl(fileno(packfile), F_SETFD, FD_CLOEXEC))
		perror("fcntl(fileno, F_SETFD, FD_CLOEXEC)");

	return errOk;
error_out:
	fprintf(stderr, __FILE__ " pakfInit(): %s invalid\n", path);
	if (packfile)
		fclose(packfile);
	packfile=NULL;
	return errOk;
}

void pakfClose(void)
{
	if (packfile)
	{
		free(dir);
		fclose(packfile);
	}
	packfile=NULL;
}
